"""
History of manipulation with containers -- create, change of state, or deletion.
"""

import logging
from collections import OrderedDict

from .constants import APP_NAME
from .output import CHAPTER

_logger = logging.getLogger(APP_NAME)


class ContainerHistory(object):
    """
    Saves the container manipulation history.
    """

    def __init__(self):
        self._hist = OrderedDict()

    def add_container_status(self, id, state, message=None):
        """
        Update the state of given container.

        * id: id of the container to save
        * state: new state of the container
        * message: message, that can be forwarded to the log
        """
        debug_text = "container {}\n- id: {}".format(state, id)
        if message:
            debug_text += "\n{}".format(message)
        _logger.debug(CHAPTER.format(debug_text))
        self._hist[id] = state

    def remove_container(self, id):
        """
        Remove container from the history.
        From now, the container is taken as deleted.
        * id: id of the container to remove
        """
        _logger.debug(CHAPTER.format("container removed\n- id: {}".format(id)))
        del self._hist[id]

    @property
    def containers(self):
        """
        * **return:** containers present in the history (containers that were not deleted)
        """
        return list(self._hist)



"""
Various utility classes used during the build.
"""

from .constants import VOLUME_MODES


class ResultImage(object):
    """
    Represent final image built by Incubator.
    """

    def __init__(self, image_id, logs=None):
        self._id = image_id
        self._logs = logs or []

    @property
    def id(self):
        """
        ID of the newly created image.
        """
        return self._id

    @property
    def logs(self):
        """
        Logs captured during the build.
        """
        return self._logs

    def __str__(self):
        return "ResultImage<id: {}>".format(str(self._id))


class CommandLog(object):
    """
    Represent logs for one dockerfile command.
    """

    def __init__(self, command, logs=None):
        self._command = command
        self._logs = logs or []

    @property
    def command(self):
        """
        Content of the instruction.
        """
        return self._command

    @property
    def logs(self):
        """
        List of string captured during the execution.
        """
        return self._logs

    def add_log(self, log):
        """
        Add new log from the executed command.

        * log: string representation of the log entry
        """
        self._logs.append(log)

    def add_logs(self, logs):
        """
        Add multiple logs capture during the execution.

        * logs: list of strings with logs
        """
        self._logs += logs


class Volume(object):
    """
    Representing a mount volume.
    """

    def __init__(self, source, destination, mode=None):
        self.source = source
        self.destination = destination
        self.mode = mode

    @staticmethod
    def get_instance(value):
        """
        Try to create a new Volume instance.

        * value: Volume instance or string in the format source:destination or source:destination:mode
        * **return:** Volume instance when the input is valid, else raises an ValueError
        """
        if not value:
            raise ValueError("None cannot be converted to the Volume instance.")

        if isinstance(value, Volume):
            return value

        value_split = value.split(':')
        if len(value_split) == 2:
            source, destination = value_split
            mode = None
        elif len(value_split) == 3:
            source, destination, mode = value_split
            if not mode:
                raise ValueError(
                    "{} is not a valid volume in format "
                    "'source:destination' or 'source:destination:mode'."
                    "('source' and 'destination' should be nonempty)".format(
                        value))
        else:
            raise ValueError(
                "{} is not a valid volume in format "
                "'source:destination' or 'source:destination:mode'.".format(value))

        if not source or not destination:
            raise ValueError(
                "{} is not a valid volume in format "
                "'source:destination' or 'source:destination:mode'."
                "('source' and 'destination' should be nonempty)".format(
                    value))

        if mode and mode not in VOLUME_MODES:
            raise ValueError("{} is not a valid mode for volume. "
                             "(valid modes: {})".format(mode, VOLUME_MODES))

        return Volume(source=source,
                      destination=destination,
                      mode=mode)

    @staticmethod
    def get_instances(value_list):
        """
        Try to get list of Volume instances from the list.

        * **return:** list of valid Volume instances or raise a ValueError
        """
        return [Volume.get_instance(v) for v in value_list] if value_list else None

    def __str__(self):
        result = "{}:{}".format(self.source, self.destination)
        if self.mode:
            result += ":{}".format(self.mode)
        return result

    def __eq__(self, o):
        if not isinstance(o, Volume):
            return False

        return self.source == o.source and self.destination == o.destination and self.mode == o.mode

    def __hash__(self):
        result = (self.source.__hash__() ** 2) * self.destination.__hash__()
        if self.mode:
            result *= self.mode.__hash__() ** 3
        return int(result)

    @staticmethod
    def volumes_to_list(volumes):
        """
        Get the list of string representation of volumes' list.

        * volumes: list with Volume instances
        * **return:** list of string representation of volumes
        """
        return [str(v) for v in volumes]

    @staticmethod
    def volumes_to_bind_dict(volumes):
        """
        Convert the volumes' list to the bind-mount dictionary format.

        * volumes: list of Volume instances
        * **return:** dictionary with bind mounts
        """
        result = {}
        for v in volumes:
            result[v.source] = {
                "bind": v.destination,
                "mode": v.mode
            }
        return result



"""
Utility methods used for files and archives manipulation and filtering.
"""
import fnmatch
import logging
import os

from .constants import APP_NAME, FILE_MODE_FROM_URL

_logger = logging.getLogger(APP_NAME)


def match_filters(file, patterns):
    """
    * file: path of the file
    * patterns: list of shell wildcard patterns
    * **return:** True, if at least one pattern satisfy the given file path
    """
    for p in patterns:
        if fnmatch.fnmatchcase(file, p):
            return True
    return False


def filter_files(files, patterns):
    """
    Filter files with given patterns.

    * files: list of file-paths
    * patterns: list of shell wildcard patterns
    * **return:** list of files, that satisfy at least one pattern
    """
    filtered_files = []

    for f in files:
        if match_filters(f, patterns):
            filtered_files.append(f)
    return filtered_files


def get_filtered_files(files, patterns, tar):
    """
    Get file-object from the archive that satisfies the given patterns.

    * files: list of file-paths
    * patterns: list of shell wildcard patterns
    * tar: tar archive
    * **return:** List of file-like objects, that satisfies the patterns
    """
    root_dirs = []

    filtered_objects = set([])
    for f in files:
        if match_filters(f, patterns):
            fi = tar.getmember(f)
            if fi.isfile():
                filtered_objects.add(f)
            elif fi.isdir():
                root_dirs.append(f)

    root_dirs.sort(key=len)
    for d in root_dirs:
        pattern = "{}{}*".format(d.split(os.sep)[-1],
                                 os.sep)
        matched_subfiles = filter_files(files, [pattern])
        if len(matched_subfiles) == 0:
            if len(d.split(os.sep)) > 1:  # empty subdirectory
                filtered_objects.add(d)
        else:
            filtered_objects.update(matched_subfiles)
            for subfile in matched_subfiles:
                if subfile in root_dirs:
                    root_dirs.remove(subfile)
    return filtered_objects


def change_tarinfo_properties(tarinfo):
    """
    Filter for insertion local file to tar archive.

    * tarinfo: tarinfo to insert
    * **return:** original tarinfo with changed permissions
    """
    tarinfo.uid = 0
    tarinfo.gid = 0
    return tarinfo


def change_tarinfo_properties_mode(tarinfo):
    """
    Filter for insertion not-local file to tar archive.

    * tarinfo: tarinfo to insert
    * **return:** original tarinfo with changed permissions
    """
    tarinfo.uid = 0
    tarinfo.gid = 0
    tarinfo.mode = FILE_MODE_FROM_URL
    return tarinfo


def normalise_patterns(patterns):
    """
    Normalize the path in the pattern.
    * patterns: list of patterns
    * **return:** list of normalized patterns
    """
    return [os.path.normpath(p) for p in patterns]


def remove_top_parent(path):
    """
    Remove the most top parent in the file path.
    """
    return os.sep.join(path.split(os.sep)[1:])


def is_top_file(path):
    """
    Test, if the path is in the top directory.
    """
    return len(path.split(os.sep)) > 1


def put_archive(client, config, container, tar, target):
    """
    Send an archive with data to the container.

    * client: containerization client to use
    * config: build configuration
    * container: id of the current container
    * tar: archive with files to send
    * target: destination path to send data in
    """
    _logger.debug('------> creating target folder:\n{}\n'.format(target))
    create_directory(client, config, container, target)
    _logger.debug('------> sending data to the container')
    client.put_archive(container=container,
                       path=target,
                       data=tar.getvalue())


def create_directory(client, config, container, target):
    """
    Create recursively directory in the container.

    * client: containerization client
    * config: build configuration
    * container: id of the current container
    * target: directory to create
    """
    client.execute(container=container,
                   cmd=config.mkdir_command + ' {}'.format(
                       target))


def copy_files_to_container(container, client, config, context, source, target):
    """
    Copy files to the container.

    * container: id of the current container
    * client: containerization client
    * config: build configuration
    * context: build context with the source files
    * source: list of source patterns
    * target: destination path
    """
    single_file = None
    if target != os.curdir and target[-1] != os.path.sep:  # destination is file
        single_file = os.path.basename(target)
        target = os.path.dirname(os.path.join(os.curdir, target))

    tar = context.get_archive(patterns=source,
                              single_file=single_file)

    if not tar:
        _logger.warning('------> nothing to copy\n{}'.format(target))
        return

    put_archive(client, config, container, tar, target)


def norm_path_with_work_dir(path, workdir):
    """
    Normalize the path relatively to the current working directory.

    * path: file path to normalize
    * workdir: current working directory
    * **return:** file-path, that was make relative to working directory and normalized
    """
    if not workdir:
        return path

    has_sep = path[-1] == os.sep
    return_path = os.path.normpath(os.path.join(workdir, path))
    if has_sep:
        return_path += os.sep
    return return_path



"""
Implementation of *Dockerfile* commands and methods for splitting the *Dockerfile* instructions to layers.
"""

import logging
import os
import tarfile

from .constants import APP_NAME
from .fileutils import (copy_files_to_container, create_directory,
                        norm_path_with_work_dir, put_archive)
from .layer import ImageLayer
from .output import CHAPTER, ERROR_SUBTITLE
from .parsing import apply_shell_expansion, substitute_variables
from .utilclasses import CommandLog

_logger = logging.getLogger(APP_NAME)


class Command:
    """
    Base class for all *Dockerfile* instructions.
    """

    def __init__(self, value, content,
                 line_envs, line_labels):
        self.value = value
        self._content = content
        self._line_envs = line_envs
        self._line_labels = line_labels

    @staticmethod
    def changes_data():
        """
        Whether the instruction changes the data in the image
        or if the instruction can be merged to the following instruction
        without commit.

        * **return:** True, if instruction changes image file-system.
        """
        pass

    @staticmethod
    def raw_value():
        """
        Whether the instruction needs the raw data, or if the value can be
        shell expanded.

        * **return:** True, if the instruction needs non-shell-expanded value.
        """
        return False

    def apply(self, client, layer_context, build_context, config):
        """
        Apply the  instruction to the  build-time container.

        * client: container client allowing the basic methods for image/container manipulation
        * layer_context: Context for the current layer, allows saving metadata to the commit configuration
            and editing the global context (user, workdir, envs).
        * build_context: allows access to the build context files
        * config: build configuration
        """
        pass

    def __str__(self):
        return "{} {}".format(self.__class__.__name__, self.value)

    def __call__(self, client, layer_context, build_context, config):
        """
        Call the instruction execution.

        * client: container client allowing the basic methods for image/container manipulation
        * layer_context: Context for the current layer, allows saving metadata to the commit configuration
            and editing the global context (user, workdir, envs).
        * build_context: allows access to the build context files
        * config: build configuration
        """
        _logger.info(CHAPTER.format(self._content))
        self.value = substitute_variables(s=self.value,
                                          variables=layer_context.global_context.envs)

        if not self.raw_value():
            self.value = apply_shell_expansion(self.value)
        self.apply(client=client,
                   layer_context=layer_context,
                   build_context=build_context,
                   config=config)


class Arg(Command):
    """
    `ARG` *Dockerfile* instruction.
    Allows predefining the build-time environments, that can be defined as a `--build-arg` argument during the build.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        value_split = self.value.split('=', 1)

        if len(value_split) == 1:
            key, default_value = value_split[0], ""
        elif len(value_split) == 2:
            key, default_value = value_split
        else:
            raise Exception("Wrong format of ARG command.")

        layer_context.use_buildarg(key=key,
                                   value=config.buildargs.get(key, default_value))


class Copy(Command):
    """
    `COPY` *Dockerfile instruction.
    Allows copying of files to the image. Cannot use remote or archive source. More time-effective than `ADD`.
    """

    @staticmethod
    def changes_data():
        return True

    def apply(self, client, layer_context, build_context, config):
        splited_command = self.value.split()
        if len(splited_command) < 2:
            _logger.warning('Wrong command: {}'.format(str(self.value)))
            return
            # exception?

        source_files = splited_command[0:-1:]
        target_path = norm_path_with_work_dir(path=splited_command[-1],
                                              workdir=layer_context.global_context.workdir)
        copy_files_to_container(container=layer_context.current_container,
                                client=client,
                                config=config,
                                context=build_context,
                                source=source_files,
                                target=target_path)


class Add(Command):
    """
    `ADD` *Dockerfile instruction.
    Allows copying of files to the image. Can use remote or archive source. Slower than `ADD`.
    The remote source is not ready yet.
    """

    @staticmethod
    def changes_data():
        return True

    def apply(self, client, layer_context, build_context, config):
        splited_command = self.value.split()
        if len(splited_command) < 2:
            _logger.warning('Wrong command: {}'.format(str(self.value)))
            return

        source_files = splited_command[0:-1:]
        target_path = norm_path_with_work_dir(path=splited_command[-1],
                                              workdir=layer_context.global_context.workdir)

        for src in source_files:
            file = build_context.get_file(src)
            if file:
                try:
                    with tarfile.open(fileobj=file):
                        put_archive(client=client,
                                    config=config,
                                    container=layer_context.current_container,
                                    tar=file,
                                    target=target_path)
                        break
                except:
                    pass
            copy_files_to_container(container=layer_context.current_container,
                                    client=client,
                                    config=config,
                                    context=build_context,
                                    source=[src],
                                    target=target_path)


class Cmd(Command):
    """
    `CMD` *Dockerfile instruction.
    Sets default arguments to the `ENTRYPOINT`.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.add_commit_property("Cmd", self.value)


class Entrypoint(Command):
    """
    `ENTRYPOINT` *Dockerfile instruction.
    Sets default command for the execution.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.add_commit_property("Entrypoint", self.value)


class Env(Command):
    """
    `ENV` *Dockerfile instruction.
    Defines the new environment variable(s).
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.update_envs(self._line_envs)


class Expose(Command):
    """
    `EXPOSE` *Dockerfile instruction.
    Predefine the port to expose.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        ports = self.value.split(" ")
        for p in ports:
            layer_context.add_commit_property_to_subdict("ExposedPorts", p, {})


class From(Command):
    """
    `FROM` *Dockerfile instruction.
    Defines the base image. Should be the first instruction in the *Dockerfile*.
    """

    @staticmethod
    def changes_data():
        return False


class Label(Command):
    """
    `LABEL` *Dockerfile instruction.
    Defines the metadata *key-value(s)* of the image.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.update_labels(self._line_labels)


class Maintainer(Command):
    """
    `MAINTAINER` *Dockerfile instruction.
    Sets the author of the *Dockerfile*/image.
    **Depracated** instruction, use `LABEL` instead.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.author = self.value


class Run(Command):
    """
    `RUN` *Dockerfile instruction.
    Executes the command in the build-time container.
    """

    @staticmethod
    def changes_data():
        return True

    @staticmethod
    def raw_value():
        return True

    def apply(self, client, layer_context, build_context, config):

        command = self.value
        log = CommandLog(command=self._content)

        if layer_context.global_context.workdir and layer_context.global_context.workdir != os.curdir:
            command = "cd {}; {}".format(layer_context.global_context.workdir, command)

        for l in client.execute(container=layer_context.current_container,
                                cmd=command,
                                envs=layer_context.global_context.envs,
                                user=layer_context.global_context.user):
            l_decode = str(l.decode().rstrip())
            _logger.debug(l_decode)
            log.add_log(l_decode)
        _logger.debug("")
        layer_context.logs.append(log)


class StopSignal(Command):
    """
    `STOPSIGNAL` *Dockerfile instruction.
    Sets the stop signal for the image.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.add_commit_property("StopSignal", self.value)


class User(Command):
    """"
    `USER` *Dockerfile instruction.
    Sets the for the following instructions and the image.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.update_user(self.value)


class Volume(Command):
    """
    `VOLUME` *Dockerfile instruction.
    Predefine the image volume.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        volumes = self.value.split(" ")
        for v in volumes:
            layer_context.add_commit_property_to_subdict("Volumes", v, {})


class Workdir(Command):
    """
    `WORKDIR` *Dockerfile instruction.
    Defines working directory for following instructions and for the image.
    """

    @staticmethod
    def changes_data():
        return False

    def apply(self, client, layer_context, build_context, config):
        layer_context.change_workdir(self.value)
        create_directory(client=client,
                         config=config,
                         container=layer_context.current_container,
                         target=layer_context.global_context.workdir)


COMMANDS = {
    "ADD": Add,
    "ARG": Arg,
    "CMD": Cmd,
    "COPY": Copy,
    "ENTRYPOINT": Entrypoint,
    "ENV": Env,
    "EXPOSE": Expose,
    "FROM": From,
    "LABEL": Label,
    "MAINTAINER": Maintainer,
    "RUN": Run,
    "STOPSIGNAL": StopSignal,
    "USER": User,
    "VOLUME": Volume,
    "WORKDIR": Workdir
}


def _get_layer_split_marks(layers, instruction_count):
    """
    Validate the layer split marks. Expand the shortcut [] to `[0,1, ... ,instruction_count]`.

    * layers: list with split marks
    * instruction_count: number of instructions
    * **return:** valid list of split marks.
    """
    layers_split_mark = layers
    if layers_split_mark is None or layers_split_mark == []:
        layers_split_mark = [i for i in range(instruction_count)]

    if layers_split_mark[-1] != instruction_count:
        layers_split_mark.append(instruction_count)

    _logger.debug(CHAPTER.format("layer splits: {}".format(layers_split_mark)))

    for l in layers_split_mark:
        if l > instruction_count:
            raise Exception("Layer split {} is bigger then number of instructions ({})."
                            .format(l, instruction_count))
    return layers_split_mark


def get_layers_from_dockerfile_structure(structure, env_label_context, layers, global_context):
    """
    Get the list of layers from the *Dockerfile* structure.

    * structure: structure of the *Dockerfile* defined by the *Dockerfile-parse*
    * env_label_context: structure of `envs`/`label` for each *Dockerfile* instruction
    * layers: list of layer split marks
    * global_context: global context (user, workdir, envs) for the whole build
    * **return:** list of layers containing the instructions
    """
    layers_split_mark = _get_layer_split_marks(layers=layers,
                                               instruction_count=len(structure))

    layers = []
    current_command = 0
    layer_number = 1
    data_changed = False

    layer = ImageLayer(number=0,
                       global_context=global_context)

    for layer_split in layers_split_mark:

        for i in range(current_command, layer_split):
            cmd = structure[i]
            context = env_label_context[i]

            instruction = cmd["instruction"]
            cmdType = COMMANDS.get(instruction)
            if not cmdType:
                _logger.warning(ERROR_SUBTITLE.format("Unknown instruction {}.".format(instruction)))
            else:
                layer.add_command(command=cmdType(value=cmd["value"],
                                                  content=cmd['content'],
                                                  line_envs=context.line_envs,
                                                  line_labels=context.line_labels),
                                  content=cmd['content'])

                current_command += 1
                data_changed = data_changed or cmdType.changes_data()

        if data_changed or layer_split == len(structure):
            data_changed = False
            layers.append(layer)
            if not layer_split == len(structure):
                layer = ImageLayer(number=layer_number,
                                   global_context=global_context)
                layer_number += 1
            else:
                layer.is_last = True

    return layers



"""
Python API for building docker images in better and more secure way.
"""

import logging

from .core import utils
from .core.build import Builder
from .core.caching import ImageCache
from .core.client import DockerClient
from .core.config import ImageConfig
from .core.constants import APP_NAME
from .core.context import BuildContext
from .core.output import ERROR_SUBTITLE
from .core.utilclasses import Volume

_logger = logging.getLogger(APP_NAME)


def build(path=None, fileobj=None, tag=None,
          custom_context=False, pull=False, rm=False,
          forcerm=False, dockerfile=None, buildargs=None,
          container_limits=None, volumes=None, layers=None,
          labels=None, config=None, context_file_limit=None,
          container_client=None, cache=None):
    """
    Build a container image from provided build context and Dockerfile.

    * path: (str) path to the directory containing *Dockerfile*
    * fileobj: a file object to use as *Dockerfile* (or a file-like
                object, e.g. `BytesIO`)
    * tag: (str) tag to add to the final image
    * custom_context: (bool) optional if using `fileobj`
    * pull: (bool) downloads any updates to the `FROM` image in *Dockerfile*
    * rm: (bool) remove intermediate containers
                The `docker build` command now defaults to `--rm=true`,
                but we have kept the old
                default of `False` to preserve backward compatibility.
    * forcerm: (bool) always remove intermediate containers,
                even after unsuccessful build
    * dockerfile: (str) path within the build context to the *Dockerfile*
    * buildargs:  (dict) dictionary of build arguments
    * container_limits: (dict) A dictionary of limits applied to each
                    container created by the build process.
                    Valid keys:
        - memory (int): set memory limit for build
        - memswap (int): total memory (memory + swap), -1 to disable swap
        - cpushares (int): CPU shares (relative weight)
        - cpusetcpus (str): CPUs in which to allow execution, e.g., `0-3`, `0,1`
    * labels: (dict) dictionary of labels to set on the image
    * volumes: (list) list of bind volumes (str or `Volume` instance)
    * layers: (list) list of layer splits
    * config: (dict or list of dicts) dictionary of addition configuration
    * context_file_limit: (int) maximum size of in memory config (0 is unlimited and default)
    * cache: object responsible for caching images
    * container_client: client representing basic container API

    * **return:** `ResultImage` instance
    """

    try:
        container_client = container_client or DockerClient()
    except Exception as ex:
        _logger.warning(ERROR_SUBTITLE.format("Cannot launch container client:\n{}".format(ex)))
        raise ex

    cache = cache or ImageCache()

    try:
        config_object = ImageConfig.merge_configs(config)
    except Exception as ex:
        _logger.warning(ERROR_SUBTITLE.format("Cannot merge config files:\n{}".format(ex)))
        raise ex

    try:
        config_object.update(buildargs=buildargs,
                             container_limits=container_limits,
                             context_file_limit=context_file_limit,
                             forcerm=forcerm,
                             labels=labels,
                             layers=layers,
                             pull=pull,
                             rm=rm,
                             tags=tag,
                             volumes=Volume.get_instances(volumes))
    except Exception as ex:
        _logger.warning(ERROR_SUBTITLE.format("Cannot update config with arguments:\n{}".format(ex)))
        raise ex

    try:
        context_fileobj, dockerfile_fileobj = utils.mkbuildcontext(path=path,
                                                                   fileobj=fileobj,
                                                                   custom_context=custom_context,
                                                                   dockerfile=dockerfile,
                                                                   client=container_client,
                                                                   limit=config_object.context_file_limit)
        build_context = BuildContext(context=context_fileobj, limit=context_file_limit)

    except Exception as ex:
        _logger.warning(ERROR_SUBTITLE.format("Cannot make build context:\n{}".format(ex)))
        raise ex

    try:
        builder = Builder(context=build_context,
                          dockerfile=dockerfile_fileobj,
                          client=container_client,
                          config=config_object,
                          cache=cache)
        return builder.build()
    except Exception as ex:
        _logger.warning(ERROR_SUBTITLE.format("Build failed:\n{}".format(ex)))
        raise ex



"""
<!---
The MIT License (MIT)

Copyright (c) 2017 Frantisek Lachman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
-->
[![build status](https://gitlab.com/lachmanfrantisek/incubator/badges/master/build.svg)
](https://gitlab.com/lachmanfrantisek/incubator/commits/master)
[![coverage report](https://gitlab.com/lachmanfrantisek/incubator/badges/master/coverage.svg)
](https://gitlab.com/lachmanfrantisek/incubator/commits/master)
[![License: MIT](https://img.shields.io/pypi/l/incubator.svg)](https://opensource.org/licenses/MIT)
[![Documentation](https://img.shields.io/badge/documentation-pdoc-green.svg)](https://lachmanfrantisek.gitlab.io/incubator/)
[![PyPI - incubator](https://img.shields.io/pypi/v/incubator.svg)](https://pypi.python.org/pypi/incubator/)
[![PyPI - incubator](https://img.shields.io/pypi/status/incubator.svg)](https://pypi.python.org/pypi/incubator/)
[![PyPI - incubator](https://img.shields.io/pypi/pyversions/incubator.svg)](https://pypi.python.org/pypi/incubator/)

---------

**Incubator is a Python API and command line interface for building container images in better and more secure way.**

With Incubator you can:

* precisely control layering of the image
* mounting build-time volumes (for secrets present only during the build)
* better metadata handling
* use standard *Dockerfile* (extended functionality is defined externally)

Source is hosted at [gitlab.com/lachmanfrantisek/incubator](https://gitlab.com/lachmanfrantisek/incubator).
"""

import logging

import os
import pkg_resources
import incubator.core.constants as _const

try:
    __version__ = pkg_resources.require(_const.APP_NAME)[0].version
except:
    _dirname = os.path.dirname(os.path.abspath(__file__))
    _version_file = open(os.path.join(_dirname, os.pardir, 'VERSION'))
    __version__ = _version_file.read().strip()


def set_logging(name=_const.APP_NAME, level=logging.DEBUG, handler=None, add_handler=True):
    """
    Set logging handler for the logger.

    * name: (str) name of the logger
    * level: logging level
    * handler: logging handler to add to logger
    * add_handler: (bool) whether to add handler to logger
    """
    logger = logging.getLogger(name)
    logger.handlers = []
    logger.setLevel(level)

    if add_handler:
        if not handler:
            handler = logging.StreamHandler()
            handler.setLevel(logging.DEBUG)

            formatter = logging.Formatter('%(message)s')
            handler.setFormatter(formatter)

        logger.addHandler(handler)



"""
Classes for representing one layer during the image build.
"""

import os

from .constants import COMMIT_MESSAGE_START


class ImageLayer(object):
    """
    Represent one layer in final image.

    Contains commands and configuration for this layer.
    """

    def __init__(self, number, global_context, commands=None):
        self._commands = commands or []
        self.number = number
        self.is_last = False
        self.commit_message = COMMIT_MESSAGE_START
        self.context = LayerContext(global_context=global_context)

    def __str__(self):
        commands = "["
        for cmd in self.commands:
            commands += "<{}>".format(str(cmd))
        commands += "]"
        return "layer {}: {}".format(str(self.number), commands)

    def add_command(self, command, content):
        """
        Add instruction to the layer and update the label for this layer.

        * command: instance of the Command
        * content: string representation of the instruction content
        """
        self._commands.append(command)

        formated_content = " {}".format(content.rstrip())
        self.commit_message += formated_content
        label_name = "layer_{}_commands".format(str(self.number))

        self.context.labels.setdefault(label_name, "")
        self.context.labels[label_name] += formated_content

    @property
    def commands(self):
        """
        Instructions on this layer.
        """
        return self._commands


class LayerContext(object):
    """
    Represent context and configuration of one layer.

    Contains metadata and commit config for new layer.
    """

    def __init__(self, global_context):
        self.global_context = global_context
        self.container = None
        self.author = None
        self.labels = {}

        self._envs = {}
        self._logs = []

        self._commit_config_properties = {}

    def add_commit_property(self, key, value):
        """
        Add property to be used for creating the commit configuration.

        * key: name of the property
        * value: new value of the  property
        """
        self._commit_config_properties[key] = value

    def add_commit_property_to_subdict(self, subdict, key, value):
        """
        Add property to be used in the directory, that will be present in the commit configuration.

        * subdict: name of the subdirectory
        * key: name of the property
        * value: new value of the property
        """
        self._commit_config_properties.setdefault(subdict, {})
        self._commit_config_properties[subdict][key] = value

    def update_labels(self, labels):
        """
        Save the labels to be used for creating the commit configuration.
        * labels: dictionary with labels
        """
        self.labels.update(labels)

    def update_envs(self, envs):
        """
        Update the global environment variables and save it for creating the commit configuration.
        * envs: New environment variables
        """
        self.global_context.envs.update(envs)
        self._envs.update(envs)

    def update_user(self, user):
        """
        Set user name or id for creating the commit configuration and update global context.

        * user: user name or number to set
        """
        self.add_commit_property("User", user)
        self.global_context.user = user

    def use_buildarg(self, key, value):
        """
        Update the build-time environment variables and mark the build argument as used.

        * key: name of the build argument
        * value: value of the build argument
        """
        self.global_context.use_buildarg(key, value)

    def add_log(self, log):
        """
        Add the log entry to logs from current layer
        * log: CommandLog instance
        """
        self._logs.append(log)

    def change_workdir(self, workdir):
        """
        Change the current working directory relatively for the old one.

        * workdir: new working directory (can be relative to old one)
        * **return:** new working directory
        """
        new_workdir = os.path.normpath(os.path.join(self.global_context.workdir, workdir))
        self.add_commit_property("WorkingDir", new_workdir)
        self.global_context.workdir = new_workdir

    @property
    def commit_config(self):
        """
        Read-only property which can be used as a commit-config dictionary
        on container commit.

        * **return:** dict as a commit-config for this layer
        """
        config = dict(self._commit_config_properties)

        config.setdefault("Env", [])
        for k, v in self._envs.items():
            config["Env"].append('{}={}'.format(str(k), str(v)))

        config.setdefault("Labels", {})
        config["Labels"].update(self.labels)

        return config

    @property
    def logs(self):
        """
        Logs from all instruction on current layer.
        """
        return self._logs



"""
Module for caching of the build images.

Nowadays, only an empty implementation.
"""

import logging

from .constants import APP_NAME
from .output import SUBTITLE

_logger = logging.getLogger(APP_NAME)


class ImageCache:
    """
    Empty implementation of the caching class.
    """

    def add(self, image_id, parent_id, layer):
        """
        Save new image to the cache.

        * image_id: id of the current image
        * parent_id: id of the parent image
        * layer: current layer (with instructions executed on the parent image to get current image)
        """
        _logger.debug(SUBTITLE.format('CACHE - SAVE'))
        _logger.debug('id: {}\nparent: {}\n{}\n'.format(str(image_id), str(parent_id), str(layer)))

    def get_layer(self, parent, layer):
        """
        Try to find image in cache.

        * parent: id of the parent image
        * layer: current layer (instructions, that has to be executed on top of the parent image)
        * **return:** id of the image, or `None`
        """
        _logger.debug(SUBTITLE.format('CACHE - IMAGE NOT FOUND'))
        _logger.debug('parent: {}\n{}\n'.format(str(parent), str(layer)))
        return None



"""
Methods for parsing, escaping and quoting the strings.
"""

import collections
import re
import shlex
import string

import dockerfile_parse
import six


def create_dockerfile_structure(dockerfile_content):
    """
    Get the structure of dockerfile of the given *Dockerfile* content.

    * dockerfile_content: file-like object with the *Dockerfile* content
    * **return:** triple:
        - command structure
        - env_label structure for each line
        - base image
    """
    dfp = dockerfile_parse.DockerfileParser(fileobj=dockerfile_content,
                                            cache_content=True,
                                            env_replace=False)

    commands = dfp.structure
    env_label_context = dfp.context_structure
    base_image = dfp.baseimage
    return commands, env_label_context, base_image


def apply_shell_expansion(str):
    """
    Use shlex to expand string. (like echo)
        e.g. "string with "space -> string with space
    * str: string to apply an expansion
    * **return:** shell-expanded string
    """
    return " ".join(shlex.split(str))


_find_unsafe = re.compile(r'[^\w@%+=:,./-]').search


def _quote_from_shlex(s):
    """
    Return a shell-escaped version of the string *s*.
    (implementation from shlex for python3)
    """
    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


def quote(s):
    """
    Quote the string to be agnostic to shell expansion.
    (Use implementation from shlex.)

    * s: string to quote
    * **return:** shell-quoted string
    """
    if six.PY3:
        return shlex.quote(s)
    else:
        return _quote_from_shlex(s)


def substitute_variables(s, variables):
    """
    Make a variable substitution.

    * s: text to apply substitution on
    * variables: (dict) environment variables
    * **return:** string s with substituted variables
    """

    s = s.replace("\\$", "$$")

    regex = "{([^}]*:-[^}]*)}"
    matches = re.finditer(regex, s)
    for m in matches:
        key, val = m.group(1).split(":-")
        variables.setdefault(key, val)

    regex_subst = "\$({[^}]*):-[^}]*}"
    s = re.sub(regex_subst, "$\g<1>}", s, 0)

    d = collections.defaultdict(lambda: "")
    d.update(variables)
    t = string.Template(s)
    substituted = t.substitute(d)
    return substituted


def get_name_and_tag(name):
    """
    Split the name if the image into tuple (name, tag).

    * name: (str) name of the image
    * **return:** tuple:
        - repository
        - tag
    """
    name_parts = name.split(':')
    if len(name_parts) == 2:
        repo = name_parts[0]
        tag = name_parts[1]
    else:
        repo = name
        tag = "latest"

    return repo, tag



"""
Utility methods for builder.
"""
import io
import logging
import os
import tarfile
import tempfile

import six

from .constants import (APP_NAME, CONTEXT_FILE_IN_MEMORY, CONTEXT_FILE_ON_DISK,
                        DEFAULT_CONTEXT_FILE_LIMIT)
from .fileutils import change_tarinfo_properties
from .output import CHAPTER

_logger = logging.getLogger(APP_NAME)


def dockerfile_from_context(context, path):
    """
    Get a *Dockerfile* file-like object from the archive with the build context.

    * context: tar archive with the build context
    * path: *Dockerfile* location in the build context
    * **return:** file-like *Dockerfile* content
    """
    with tarfile.open(mode="r", fileobj=context) as t:
        df = t.extractfile(t.getmember(path))
    dockerfile = six.BytesIO()
    dockerfile.writelines(df.readlines())
    dockerfile.seek(0)
    context.seek(0)
    return dockerfile


def mkbuildcontext(path, fileobj, custom_context, dockerfile, client, limit=None):
    """
    Make a build context from the variety of arguments.

    * path: file-path to the build context
    * fileobj: file-like object representing *Dockerfile* or build context with *Dockerfile*
    * custom_context: whether to use custom context
    * dockerfile: *Dockerfile* location in the build context
    * client: containerization client
    * limit: limit of context size to save context in memory
    * **return:** tuple:
        - context file-like object
        - *Dockerfile* file-like object
    """
    if path is None and fileobj is None:
        raise TypeError("Either path or fileobj needs to be provided.")

    dockerfile = dockerfile or "Dockerfile"

    if custom_context:
        if not fileobj:
            raise TypeError("You must specify fileobj with custom_context")
        return fileobj, dockerfile_from_context(fileobj, dockerfile)
    elif fileobj is not None:
        return make_build_context_from_dockerfile(fileobj), fileobj
    elif path.startswith(('http://', 'https://',
                          'git://', 'github.com/', 'git@')):
        # TODO
        return None, None
    elif not os.path.isdir(path):
        raise TypeError("You must specify a directory to build in path")
    else:
        dockerignore = os.path.join(path, '.dockerignore')
        exclude = None
        if os.path.exists(dockerignore):
            with open(dockerignore, 'r') as f:
                exclude = list(filter(bool, f.read().splitlines()))
        context = tar(
            path, client, exclude=exclude, dockerfile=dockerfile, limit=limit
        )
        dockerfile_obj = dockerfile_from_context(context=context, path=dockerfile)
        return context, dockerfile_obj


def exclude_paths(root, exclude, dockerfile, client):
    """
    Exclude paths from the build context.

    * root: root file path
    * exclude: exclude patterns
    * dockerfile: *Dockerfile* file-path in the context
    * client: containerization context
    * **return:** filtered files
    """
    files = client.exclude_paths_from_context(root, exclude, dockerfile)
    return files


def tar(path, client, exclude=None, dockerfile=None, fileobj=None, limit=DEFAULT_CONTEXT_FILE_LIMIT):
    """
    Make an archive with the build context.

    * path: to build context
    * client: containerization client
    * exclude: exclude patterns
    * dockerfile: of the *Dockerfile* in the build context
    * fileobj: file-object representation of the *Dockerfile* or build context with the *Dockerfile*
    * limit: limit of context size to save context in memory
    * **return:** fil-like object with the build context
    """
    if not fileobj:
        fileobj = tempfile.SpooledTemporaryFile(max_size=limit)

    root = os.path.abspath(path)
    exclude = exclude or []

    _logger.debug(CHAPTER.format("context:"))
    with tarfile.open(mode='w', fileobj=fileobj) as t:
        for p in sorted(exclude_paths(root, exclude,
                                      dockerfile=dockerfile,
                                      client=client)):
            _logger.debug("- {}".format(p))
            t.add(name=os.path.join(path, p), arcname=p, filter=change_tarinfo_properties)
    _logger.debug("")
    fileobj.seek(0)

    if isinstance(fileobj, tempfile.SpooledTemporaryFile):
        if fileobj.name:
            context_file_type = CONTEXT_FILE_ON_DISK
        else:
            context_file_type = CONTEXT_FILE_IN_MEMORY
        _logger.debug(CHAPTER.format("context type: {}".format(context_file_type)))

    return fileobj


def make_build_context_from_dockerfile(dockerfile):
    """
    Make a build context file object from the provided *Dockerfile*.

    * dockerfile: file-like object with the *Dockerfile* content
    * **return:** file-like object with the build context
    """
    f = six.BytesIO()
    with tarfile.open(mode='w', fileobj=f) as t:
        if isinstance(dockerfile, six.StringIO) or isinstance(dockerfile, io.StringIO):
            dfinfo = tarfile.TarInfo('Dockerfile')
            dfinfo.size = len(dockerfile.getvalue())
            dockerfile.seek(0)
        elif isinstance(dockerfile, six.BytesIO) or isinstance(dockerfile, io.BytesIO):
            dfinfo = tarfile.TarInfo('Dockerfile')
            dfinfo.size = len(dockerfile.getvalue())
            dockerfile.seek(0)
        else:
            dfinfo = t.gettarinfo(fileobj=dockerfile, arcname='Dockerfile')
        t.addfile(dfinfo, dockerfile)
    f.seek(0)
    return f


def get_list_from_tuple_or_string(value):
    """
    Make a list of strings from given value.

    * value: None, string or list of strings
    * **return:** list with strings
    """
    if not value:
        return []
    if isinstance(value, six.string_types):
        return [value]
    else:
        return list(value)



"""
Classes for manipulating with the container engine.
"""

import ast
import logging
import re
from distutils.version import StrictVersion
from time import sleep

import docker
from docker.errors import APIError
from docker.utils import exclude_paths

from .constants import (APP_NAME, DEFAULT_INFINITE_COMMAND, MAX_API_ATTEMPTS,
                        TIME_TO_NEXT_ATTEMPT)
from .history import ContainerHistory
from .output import CHAPTER, ERROR_SUBTITLE
from .parsing import get_name_and_tag, quote
from .utilclasses import Volume

_IS_OLD_API = StrictVersion(docker.__version__) < StrictVersion('2.0.0')

_logger = logging.getLogger(APP_NAME)


class Client:
    """
    Abstract container client representing basic functionality of container engine.
    """

    def __init__(self, history=None):
        self.history = history or ContainerHistory()

    def build_image(self, fileobj=None, dockerfile=None, custom_context=None, buildargs=None, path=None):
        """
        Build image from given dockerfile, context and build arguments.

        * fileobj: file-like object representing dockerfile or archive with context
        * dockerfile: path to dockerfile in build context
        * custom_context: whether to use fileobj as context
        * buildargs: build arguments passed to build
        * path: directory with build context
        * **return:** id of created image
        """
        pass

    def run_container_infinitely(self, image, infinite_command=None, volumes=None, container_limits=None):
        """
        Run a container with infinite command.
        Container is created, never stop and can be changed via exec.

        * image: image used for new container
        * infinite_command: infinite_command can be overwritten (default is in constants.py)
        * volumes: volumes mounted to container
        * container_limits: (dict) A dictionary of limits applied to each container created by the build process. Valid keys:
                    memory (int): set memory limit for build
                    memswap (int): Total memory (memory + swap), -1 to disable swap
                    cpushares (int): CPU shares (relative weight)
                    cpusetcpus (str): CPUs in which to allow execution, e.g., "0-3", "0,1"
        * **return:** id of created container
        """
        pass

    def run_container(self, image, command):
        """
        Run a container with givent command.

        * image: container image to run
        * command: command to run in container
        * **return:** output from command
        """
        pass

    def commit_container(self, container, conf, author, message, tags=None):
        """
        Create new image from container.

        * container: id of container to commit
        * conf: configuration forced to client
        * author: author of commit
        * message: commit message
        * tags: (list of tuples) tags connected to new image
        * **return:** id of created image
        """
        pass

    def execute(self, container, cmd, envs, user):
        """
        Create new image from container.

        * container: id of container to commit
        * conf: configuration forced to client
        * author: author of commit
        * message: commit message
        * tags: (list of tuples) tags connected to new image
        * **return:** id of created image
        """
        pass

    def put_archive(self, container, path, data):
        """
        Put archive into container.

        * container: id of container to put data in
        * path: where to put data
        * data: archive with data
        * **return:** None
        """
        pass

    def remove_container(self, container):
        """
        Remove container.

        * container: id of container to remove
        * **return:** None
        """
        pass

    def remove_image(self, image):
        """
        Remove image.

        * image: Id of image to remove.
        * **return:** None
        """
        pass

    def get_image(self, name):
        """
        Get id of image with given name.

        * name: name or id of image
        * **return:** id of image with given name or None
        """
        pass

    def pull_image(self, name):
        """
        Pull image with given name from repository.

        * name: Name of image to pull. Default tag is 'latest'.
        * **return:** Id of pulled image or None.
        """
        pass

    def get_image_info(self, image):
        """
        Return information about the image (like `docker inspect image_id`

        * image: id or name of the image
        * **return:** dict with information about the given image
        """

    @staticmethod
    def exclude_paths_from_context(root, exclude, dockerfile):
        """
        * root: path of build context
        * exclude: list of `.dockerignore` patterns
        * dockerfile: path to dockerfile
        * **return:** iterator of all paths (files and directories) not excluded with patterns
                    All paths returned are relative to the root.
        """
        pass


class DockerClient(Client):
    """
    Container client representing docker-py independently to version of *Docker-py*.
    """

    def __init__(self, history=None):
        Client.__init__(self, history)
        self.client = docker.from_env(version='auto')

    def build_image(self, fileobj=None, dockerfile=None, custom_context=None, buildargs=None, path=None):
        if _IS_OLD_API:
            resp = [line for line in self.client.build(rm=True, fileobj=fileobj,
                                                       dockerfile=dockerfile,
                                                       custom_context=custom_context,
                                                       buildargs=buildargs,
                                                       path=path
                                                       )]
            event = ast.literal_eval(resp[-1].decode())
            if 'stream' in event:
                match = re.search(r'Successfully built ([0-9a-f]+)',
                                  event.get('stream', ''))
                if match:
                    image_id = match.group(1)
                    return image_id
        else:
            image = self.client.images.build(rm=True, fileobj=fileobj,
                                             dockerfile=dockerfile,
                                             custom_context=custom_context,
                                             buildargs=buildargs,
                                             path=path)
            return image.id

    def run_container_infinitely(self, image, infinite_command=None, volumes=None, container_limits=None):
        infinite_command = infinite_command or DEFAULT_INFINITE_COMMAND
        volumes = volumes or {}
        container_limits = container_limits or {}

        if _IS_OLD_API:
            client = self.client
        else:
            client = self.client.api

        volumes_list = Volume.volumes_to_list(volumes)
        container_id = client.create_container(image=image,
                                               command=infinite_command,
                                               host_config={"binds": volumes_list,
                                                            "Memory": container_limits.get("memory"),
                                                            'MemorySwap': container_limits.get("memswap"),
                                                            'CpuShares': container_limits.get("cpushares"),
                                                            'CpusetCpus': container_limits.get("cpusetcpus")
                                                            },
                                               stdin_open=True).get('Id')
        self.history.add_container_status(id=container_id,
                                          state="created")
        client.start(container_id)
        self.history.add_container_status(id=container_id,
                                          state="running")
        return container_id

    def run_container(self, image, command, remove=False, volumes=None, container_limits=None):
        volumes = volumes or {}
        container_limits = container_limits or {}
        if _IS_OLD_API:
            volumes_list = Volume.volumes_to_list(volumes)
            container_id = self.client.create_container(image=image,
                                                        command=command,
                                                        host_config={"binds": volumes_list,
                                                                     "Memory": container_limits.get("memory"),
                                                                     'MemorySwap': container_limits.get("memswap"),
                                                                     'CpuShares': container_limits.get("cpushares"),
                                                                     'CpusetCpus': container_limits.get("cpusetcpus")
                                                                     },
                                                        ).get('Id')
            self.history.add_container_status(id=container_id,
                                              state="created")
            self.client.start(container_id)
            self.history.add_container_status(id=container_id,
                                              state="running")

            # TODO: temporary workaround !!!
            self.client.wait(container_id)

            for l in self.client.logs(container=container_id,
                                      stream=True, follow=True,
                                      stdout=True, stderr=True,
                                      tail='all'):
                yield l

            self.history.add_container_status(id=container_id,
                                              state="stopped")
            if remove:
                self.client.remove_container(container_id, force=True)
                self.history.remove_container(id=container_id)

        else:
            volumes_list = Volume.volumes_to_bind_dict(volumes)
            container = self.client.containers.create(image=image,
                                                      command=command,
                                                      volumes=volumes_list,
                                                      mem_limit=container_limits.get("memory"),
                                                      memswap_limit=container_limits.get("memswap"),
                                                      cpu_shares=container_limits.get("cpushares"),
                                                      cpuset_cpus=container_limits.get("cpusetcpus"))
            self.history.add_container_status(id=container.id,
                                              state=container.status)

            container.start()

            self.history.add_container_status(id=container.id,
                                              state=container.status)

            # TODO: temporary workaround !!!
            container.wait()

            for l in container.logs(stream=True, follow=True, stdout=True, stderr=True, tail='all'):
                yield l

            self.history.add_container_status(id=container.id,
                                              state="stopped")

            if remove:
                container.remove(force=True)
                self.history.remove_container(container.id)

    def commit_container(self, container, conf, author, message, tags=None):
        tags = tags or []
        if _IS_OLD_API:
            image_id = self.client.commit(container=container,
                                          conf=conf,
                                          author=author,
                                          message=message).get('Id')
            self.history.add_container_status(container, "committed", "- image: {}".format(image_id))

            for repo, tag in tags:
                self.client.tag(image=image_id,
                                repository=repo,
                                tag=tag)
                _logger.debug(
                    CHAPTER.format("image tagged:\n- id: {}\n- repo: {}\n- tag: {}".format(image_id, repo, tag)))
            return image_id
        else:
            container_obj = self.client.containers.get(container)
            image = container_obj.commit(conf=conf,
                                         author=author,
                                         message=message)
            self.history.add_container_status(container, "committed", "- image: {}".format(image.id))
            for repo, tag in tags:
                image.tag(repo, tag)
                _logger.debug(
                    CHAPTER.format("image tagged:\n- id: {}\n- repo: {}\n- tag: {}".format(image.id, repo, tag)))
            return image.id

    def execute(self, container, cmd, envs=None, user=None):
        envs = envs or {}

        if _IS_OLD_API:
            env_string = " ".join(["{}={}".format(k, quote(v)) for k, v in envs.items()])
            cmd = "{} {}".format(env_string, cmd)

        cmd = "sh -c {}".format(quote(cmd))

        if _IS_OLD_API:
            command_id = self.client.exec_create(container=container,
                                                 cmd=cmd,
                                                 user=user).get('Id')
            output = self.client.exec_start(exec_id=command_id,
                                            stream=True)
        else:
            container_obj = self.client.containers.get(container)
            output = container_obj.exec_run(cmd=cmd,
                                            stream=True,
                                            environment=envs,
                                            user=user)
        return output

    def put_archive(self, container, path, data):
        for i in range(MAX_API_ATTEMPTS):
            try:
                if _IS_OLD_API:
                    self.client.put_archive(container=container,
                                            path=path,
                                            data=data)
                else:
                    container_obj = self.client.containers.get(container)
                    container_obj.put_archive(path=path, data=data)
            except Exception as exc:
                _logger.warning(ERROR_SUBTITLE.format("fail to send data to the container (repeating)"))
                if i == MAX_API_ATTEMPTS:
                    raise exc
                sleep(TIME_TO_NEXT_ATTEMPT)

    def stop_container(self, container):
        if _IS_OLD_API:
            self.client.kill(container)
            self.history.add_container_status(container, "stopped")
        else:
            container_obj = self.client.containers.get(container)
            container_obj.kill()
            self.history.add_container_status(container, "stopped")

    def remove_container(self, container):
        if _IS_OLD_API:
            self.client.remove_container(container)
            self.history.remove_container(container)
        else:
            container_obj = self.client.containers.get(container)
            container_obj.remove()
            self.history.remove_container(container)

    def remove_image(self, image):
        if _IS_OLD_API:
            self.client.remove_image(image)
        else:
            image_obj = self.client.images.get(image)
            image_obj.remove()

    def get_image(self, name):
        try:
            if _IS_OLD_API:
                image = self.client.inspect_image(name)
                if image:
                    return image.get('Id')
            else:
                image = self.client.images.get(name)
                if image:
                    return image.id
        except APIError:
            return None
        return None

    def pull_image(self, name):
        image_name, image_tag = get_name_and_tag(name)

        if _IS_OLD_API:
            self.client.pull(repository=image_name,
                             tag=image_tag)
            return self.get_image(name)
        else:
            image = self.client.images.pull("{}:{}".format(image_name, image_tag))
            if image:
                return image.id
        return None

    def get_image_info(self, image):
        try:
            if _IS_OLD_API:
                return self.client.inspect_image(image)
            else:
                image = self.client.images.get(image)
                return image.attrs
        except Exception:
            return None

    @staticmethod
    def exclude_paths_from_context(root, exclude, dockerfile):
        return exclude_paths(root, exclude, dockerfile)



"""
Custom Click types for Incubator.
"""
import click

from ..core.utilclasses import Volume


class KeyValue(click.ParamType):
    """
    Key-value pair in format `KEY=VALUE`.
    """
    name = 'key=value'

    def convert(self, value, param, ctx):
        value_split = value.split('=')
        if len(value_split) != 2:
            self.fail("{} is not a valid 'key=value'.".format(value), param, ctx)
        key, value = value_split
        if not key:
            self.fail("{} is not a valid 'key=value'. (key is empty)".format(value), param, ctx)
        if not value:
            self.fail("{} is not a valid 'key=value'. (value is empty)".format(value), param, ctx)
        return key, value


class MountVolume(click.ParamType):
    """
    Volumes mapping in format `SOURCE:DESTINATION` or `SOURCE:DESTINATION:MODE`,
    where `MODE` is one of `ro`, `rw`, `z`, `Z`.
    """
    name = 'volume'

    def convert(self, value, param, ctx):
        try:
            return Volume.get_instance(value)
        except ValueError as ex:
            self.fail(ex, param, ctx)


KEY_VALUE = KeyValue()
MOUNT_VOLUME = MountVolume()



"""
Core functionality for building container images.
"""

import logging

from .commands import get_layers_from_dockerfile_structure
from .constants import APP_NAME
from .context import GlobalContext
from .output import CHAPTER, ERROR_SUBTITLE, SUBTITLE, TITLE
from .parsing import create_dockerfile_structure
from .utilclasses import ResultImage

_logger = logging.getLogger(APP_NAME)


class Builder:
    """
    Class allowing the image build process.
    """

    def __init__(self, context, dockerfile, client, config, cache):
        """
        * context: build context that will be loaded during the build
                    and that can be used as a source for `ADD` and `COPY`.
        * dockerfile: file-like object with the *Dockerfile* content
        * client: container client
                    (allowing the basic operation with images and containers)
        * config: instance of *Config* class
        * cache: object with cache ability
        """

        self.client = client

        self._context = context
        self._config = config
        self._cache = cache
        self._dockerfile = dockerfile
        self._cmd_structure, self._env_label_structure, self._base_image = create_dockerfile_structure(
            dockerfile_content=self._dockerfile)

    def build(self):
        """
        Build an image from given *Dockerfile*, context and configuration.
        Using given container client.

        * **return:** id of final image
        """

        logs = []
        last_image = None
        current_layer = None
        global_context = GlobalContext(buildargs=self._config.buildargs)
        layers = get_layers_from_dockerfile_structure(structure=self._cmd_structure,
                                                      env_label_context=self._env_label_structure,
                                                      layers=self._config.layers,
                                                      global_context=global_context)

        self._prebuild()

        try:
            for current_layer in layers:
                _logger.info(TITLE.format(" LAYER {} ".format(current_layer.number)))
                cached_image = self._cache.get_layer(last_image, current_layer)
                if cached_image:
                    last_image = cached_image
                else:
                    self._create_new_layer(layer=current_layer,
                                           image=last_image)
                    for cmd in current_layer.commands:
                        cmd(client=self.client,
                            layer_context=current_layer.context,
                            build_context=self._context,
                            config=self._config)

                    if not current_layer.is_last:
                        last_image, _logs = self._middle_commit(layer=current_layer,
                                                                last_image=last_image)
                        logs += _logs

            last_image, _logs = self._final_commit(layer=layers[-1],
                                                   last_image=last_image)
            logs += _logs

        except Exception as ex:
            if self._config.forcerm:
                last_container = current_layer.context.current_container
                if last_container in self.client.history.containers:
                    self.client.remove_container(container=last_container)
            raise ex

        self._postbuild()

        if len(global_context.unused_buildargs) > 0:
            b_args = "\n- ".join(self._config.unused_buildargs)
            _logger.warning(ERROR_SUBTITLE.format("Not used build arguments:\n- {}".format(b_args)))

        return ResultImage(image_id=last_image, logs=logs)

    def _create_base_container(self):
        """
        Create a base container filled in *Dockerfile*.

        * **return:** id of a base container
        """
        _logger.debug(CHAPTER.format("base image: {}".format(self._base_image)))

        if self._config.pull or self.client.get_image(self._base_image) is None:
            self.client.pull_image(self._base_image)

        return self.client.run_container_infinitely(
            image=self._base_image,
            infinite_command=self._config.infinite_command,
            volumes=self._config.volumes,
            container_limits=self._config.container_limits)

    def _create_new_layer(self, layer, image):
        """
        Run a new container for the current layer.

        * layer: current `ImageLayer` instance
        * image: image to use for the container
                If not present, the base image from *Dockerfile* is used.
        * **return:** id of newly created container for the current layer
        """
        if not image:
            layer.context.current_container = self._create_base_container()
        else:
            layer.context.current_container = self.client.run_container_infinitely(
                image=image,
                infinite_command=self._config.infinite_command,
                volumes=self._config.volumes,
                container_limits=self._config.container_limits)

    def _final_commit(self, layer, last_image):
        """
        Run a final commit of the image and tag it with the tags from the configuration.

        * layer: current `ImageLayer` instance
        * last_image: id of the last image
        * **return:** tuple (id of the committed image and logs)
        """
        self.client.stop_container(layer.context.current_container)
        _logger.debug(TITLE.format(" FINAL COMMIT "))
        layer.context.update_labels(self._config.labels)
        image = self.client.commit_container(layer.context.current_container,
                                             conf=layer.context.commit_config,
                                             author=layer.context.author,
                                             message=layer.commit_message,
                                             tags=self._config.tags_and_repos)
        self._cache.add(image_id=image,
                        parent_id=last_image,
                        layer=layer)

        last_image = image
        if self._config.rm:
            self.client.remove_container(layer.context.current_container)

        return last_image, layer.context.logs

    def _middle_commit(self, layer, last_image):
        """
        Provide the commit between the two layers.

        * layer: current `ImageLayer` instance
        * last_image: id of the last used image
        * **return:** tuple (id of the committed image and logs)
        """
        self.client.stop_container(layer.context.current_container)
        _logger.debug(TITLE.format(" COMMIT "))
        image = self.client.commit_container(layer.context.current_container,
                                             conf=layer.context.commit_config,
                                             author=layer.context.author,
                                             message=layer.commit_message)

        self._cache.add(image_id=image,
                        parent_id=last_image,
                        layer=layer)

        last_image = image

        if self._config.rm:
            self.client.remove_container(layer.context.current_container)
        return last_image, layer.context.logs

    def _prebuild(self):
        """
        For the future use.
        """
        _logger.debug(SUBTITLE.format('prebuild'))

    def _postbuild(self):
        """
        For the future use.
        """
        _logger.debug(SUBTITLE.format('postbuild'))
